export interface IMsgHandler {
    /**
     * Handle message. If message is a query, then return result if available immediately,
     * or a promise resolved when the answer becomes available, else return undefined.
     * Throws if error occurs during processing.
     */
    handle(msg: WSMsg): any;
}
/**
 * Interface to be implemented by receivers of property/stream data obtained through
 * a PubSubPeer subscription.
 */
export interface DataHandler<T> {
    dataReceived(newvalue: T, path: string): void;
}
/**
 * Manage the client side of web socket property pub-sub communications.
 */
export declare class PubSubPeer {
    private static kStreamPrefix;
    static kChannelStreamInfix: string;
    private wsUrl;
    private readonly msgHandler;
    private readonly autoConnect;
    private subscribers;
    private toSubscribe;
    private toUnsubscribe;
    private toSet;
    private toAdd;
    private toTell;
    private lastQueryId;
    private pendingQueries;
    private socket;
    private online;
    private pendingServerCall;
    private pendingConnect;
    private readonly stateChangeNotifier;
    /**
     * Get me going, setting an optional function to be called when my
     * connection state changes. If you specify serverPath, this should
     * NOT include protocol and location.host. If you care
     * about other messages than pub-sub ones, pass in your msgHandler.
     * By default, I will attempt to connect and re-connect automatically.
     * If you don't like this, pass false to autConnect, and call
     * connect from outside (then you must also handle re-connection).
     */
    constructor(stateChangeNotifier?: (online: boolean) => void, serverPath?: string, // Default provided below
    msgHandler?: IMsgHandler, // Handles non-subscription related messages
    autoConnect?: boolean);
    /**
     * Alternative way of connecting for use if you passed autoConnect false
     * to the ctor
     */
    connect(serverPath?: string): void;
    /**
     * Subscribe to specfied path, calling dataCallback when server reports
     * a change. I return any existing value already obtained from this property,
     * if any, which caller may use. I decided to return this instead of
     * iummediately turning around and dataCallback with any preexisting data
     * as doing so reduced the chance of unexpected behavior in the callback;
     * e.g. if subscribe is called as a result of an Ancular watch firing,
     * and the callback then performing $apply, which may result in digest
     * errors, etc. This design is debatable, but was considered "safer" for
     * now since I initially wanted this preexisting data mainly for buttons
     * when a panel reconnects.
     */
    subscribe<T>(path: string, dataCallback: DataHandler<T>): T | undefined;
    /**
     * Detach dataCallback from path so it will no longer be called on changes.
     */
    unsubscribe(path: string, dataCallback: DataHandler<any>): void;
    /**
     * Convenience combination of the two above, when subscribe/unsubscribe is determined
     * by a boolean.
     */
    subUnsub<T>(subscribe: boolean, path: string, changeCallback: DataHandler<T>): void;
    /**
     * Set path to value.
     */
    set(path: string, value: any): void;
    /**
     * Add/append value to what's already at path.
     */
    add(path: string, value: any): void;
    /**
     * General "send message to endpoint". ToDo: Replace this with
     * function-mapped RPC to neuron instead of mapping this
     * through a "property", which it really isn't.
     */
    tell(path: string, value: any): void;
    /**
     * Send toSend to stream/channel with name compName, which must start with "$Stream."
     * followed by the script context (e.g., "Script.user.MyTest"), possibly followed
     * by ".$Channel." and finally the local channel name, as produced by
     * PubSubPeer.channelName()
     */
    send(compName: string, toSend: string): void;
    /**
     * Given a scriptContext (e.g. "Script.user.MyTest") and leaf channelName,
     * returns the full composite channel name, that can subsequently
     * be passed to subscribe, unsubscribe and sendStream.
     */
    static channelName(scriptContext: string, channelName: string): string;
    /**
     * Send reset command for stream at path. Initially devised to clear
     * any backlog held in the stream.
     */
    resetStream(path: string): void;
    /**
     * Send a "warning" message to the server from this spot. The server will log
     * this along with param (which is typically a Str1Par or a Str2Par.
     */
    logWarning(param: WSMsgPar): void;
    /**
     * Send an "error" message to the server from this spot. The server will log
     * this along with param (which is typically a Str1Par or a Str2Par.
     */
    logError(param: WSMsgPar): void;
    /**
     * Send msg to server ASAP.
     */
    sendMsg(msg: WSMsg): void;
    /**
     * Send msg as a question. Once the corresponding reply arrives,
     * resolve the returned promise (or reject it if we fail). Will
     * be rejected IMMEDIATELY if I'm currently offline, so you may
     * want to pre-flight that.
     *
     * ToDo: Possibly add some mechanism to reject unanswered
     * queries after some reasonably long time.
     */
    ask<ReplyT>(msg: WSMsg): Promise<ReplyT>;
    /**
     * Try to connect to web socket after delayMs.
     */
    private attemptToConnectSoon;
    /**
     * Attempt a server connection. Called through attemptToConnectSoon.
     * If fails, will request a new attempt through attemptToConnectSoon.
     */
    private attemptConnect;
    /**
     Disconnect handlers from socket and discard it. Does nothing if got no
     socket connected.
     */
    private discardSocket;
    /**
     * Set my online state and notify any listener.
     */
    private setOnline;
    /**
     * Return my online status, which may initially be undefined.
     */
    isOnline(): boolean;
    /**
     * Got what an array of messages from server. Handle protocol level commands,
     * and pass on pub-sub and other "unknown" messages to handleMsg.
     */
    private treatIncomingData;
    /**
     handle new data, of either property or stream type.
     */
    private handleNewData;
    /**
     * Got reply to a query with specified id. Resolve associated promise
     * with provided data (which hopefully is of the expected type).
     */
    private handleReply;
    /**
     * Got an ERROR reply to a query with specified id. Reject associated promise.
     */
    private handleQueryError;
    /**
     * Handle a message (except "protocol level" msgs handled above).
     */
    private handleMsg;
    /**
     * Copy all from subscribed to toSubscribe to make sure server knows
     * about this. This is done once after successfully connected, to
     * reconnect all existing subscriptions, which then likely have been
     * dropped by the server.
     */
    private subscribeAll;
    /**
     * Make sure accumulated data is sent to server ASAP. I'll schedule a call soon
     * if one isn't already scheduled. I will NOT do anything if socket not connected,
     * as the initial connection will send accumulated data.
     */
    private requestServerCall;
    /**
     * Send accumulated, if any, to server over websocket.
     */
    private tellServer;
}
/**
 * A message sent over websock connection. Also exists server side with same name.
 */
export interface WSMsg {
    name: string;
    id?: number;
    isResponse?: boolean;
    param?: WSMsgPar;
}
/**
 * Data sent with the message.
 */
interface WSMsgPar {
    type: string;
}
/**
 * A boolean command parameter.
 */
export declare class BoolPar implements WSMsgPar {
    bool: boolean;
    type: string;
    constructor(bool: boolean);
}
/**
 * A numeric command parameter.
 */
export declare class NumPar implements WSMsgPar {
    value: number;
    type: string;
    constructor(value: number);
}
/**
 * A single string command parameter.
 */
export declare class Str1Par implements WSMsgPar {
    s1: string;
    type: string;
    constructor(s1: string);
}
/**
 * A two string command parameter.
 */
export declare class Str2Par implements WSMsgPar {
    s1: string;
    s2: string;
    type: string;
    constructor(s1: string, s2: string);
}
/**
 * The "tick" message parameter, telling me about the server time, which
 * can subsequently be used to time other events with reasonable
 * accuracy.
 */
export declare class TickPar implements WSMsgPar {
    sessionId: number;
    serverTime: number;
    type: string;
    constructor(sessionId: number, serverTime: number);
}
/**
 * Synchronization message, sent to server to inform him about time position
 * of a synchronizer master. Also exists on server side under same name.
 */
export declare class SyncPar implements WSMsgPar {
    private path;
    private time;
    private rate;
    private serverTime;
    private optBlock?;
    type: string;
    constructor(path: string, // Sub-path to property under neuron holding sync data
    time: number, // Current time position
    rate: number, // Current rate (0 is pause, 1 is nominal)
    serverTime: number, // Corresponding server time (0 if unknown)
    optBlock?: string);
}
/**
 * A general purpose "object array" parameter type. Mainly for primitive
 * types, such as string, number, boolean, which works well on both sides
 * of the fence.
 */
export declare class ArrayPar implements WSMsgPar {
    elems: any[];
    type: string;
    constructor(elems: any[]);
}
/**
 * A somewhat degenerate "reply" parameter, used with my incoming "reply"
 * message only (which is never sent to server).
 */
export declare class ReplyPar<ReplyType> implements WSMsgPar {
    type: string;
    data: ReplyType;
}
/**
 * Data indicating a time position and rate. Used for playback and sync purposes.
 * Same exists server side with same name.
 */
export interface TimeFlow {
    position: number;
    rate: number;
    serverTime?: number;
    dead?: boolean;
    optBlock?: string;
}
export {};
