import SignaturePad from "./signaturePad/signature_pad";

export class PixiSignaturePad {

	private signaturePad: SignaturePad;
	private canvas = document.querySelector<HTMLCanvasElement>("#canvas");
	private canvasContext = this.canvas.getContext("2d");
	private actions = document.querySelector('#main .actions');
	private clearButton = document.querySelector('.button.clear');
	private doneButton = <HTMLElement>document.querySelector('.button.done');

	constructor(
		private signaturePadConfig = {
			minWidth: 12, // When using pen, go with 4
			maxWidth: 12, // When using pen, go with 4
			color: '#FFFFFF',
			backgroundColor: 'rgba(0, 0, 0, 0)'
		}
	) {
		this.signaturePad = new SignaturePad(
			this.canvas,
			{
				minWidth: this.signaturePadConfig.minWidth,
				maxWidth: this.signaturePadConfig.maxWidth,
				penColor: this.signaturePadConfig.color,
				backgroundColor: this.signaturePadConfig.backgroundColor,
				onEnd: this.firstLineDraw.bind(this),
				dotSize: 12 // When using pen, go with 1
			}
		);

		this.clearButton.addEventListener('pointerup', this.clear.bind(this));
		this.doneButton.addEventListener('pointerup', this.allDone.bind(this));

		this.init();
	}

	private firstLineDraw() {
		this.actions.classList.remove('disabled');
	}

	private enterWriteMode() {
		this.canvasContext.globalCompositeOperation = 'source-over';
		this.signaturePad.minWidth = this.signaturePadConfig.minWidth;
		this.signaturePad.maxWidth = this.signaturePadConfig.maxWidth;
	}

	private clear() {
		this.signaturePad.clear();
		this.actions.classList.add('disabled');
		this.enterWriteMode();
	}

	private async allDone() {
		this.actions.classList.remove('disabled');

		const resultCanvas = document.createElement('canvas');
		resultCanvas.width = 1080;
		resultCanvas.height = 1920;
		const resultContext = resultCanvas.getContext('2d');

		await this.renderSignatureToCanvas(resultContext);

		resultCanvas.toBlob((file: File) => {
			let formData = new FormData();
			formData.append("file", file);
			formData.append("target", "upload");
			formData.append("fileName", "*.png");

			let xhr = new XMLHttpRequest();
			xhr.onreadystatechange = () => {
				if (xhr.readyState === 4 && xhr.status === 200) {
					let tempPath = xhr.responseText;
					xhr = new XMLHttpRequest();
					xhr.onreadystatechange = () => {
						if (xhr.readyState === 4 && xhr.status === 200) {
							this.clear();
						}
					};
					xhr.open("POST", "/rest/script/invoke/Scribble/finish");
					xhr.setRequestHeader('Content-Type', 'application/json');
					xhr.send(JSON.stringify({
						tempPath: tempPath
					}));
				}
			};
			xhr.open("POST", "/rest/upload/multipart");
			xhr.send(formData);
		}, 'image/png', 1);
	}

	/**
	 * Renders the signature onto the result canvas.
	 */
	private renderSignatureToCanvas(context: CanvasRenderingContext2D): Promise<void> {
		return new Promise((resolve) => {
			const signatureDataURL = this.signaturePad.toDataURL();
			const imageEl = new Image();
			imageEl.width = this.canvas.width;
			imageEl.height = this.canvas.height;

			imageEl.addEventListener('load', () => {
				context.drawImage(
					imageEl,
					0,
					0,
					imageEl.width,
					imageEl.height,
					0,
					0,
					1080,
					1920
				);
				resolve();
			});
			imageEl.src = signatureDataURL;
		});
	}

	private init() {
		const canvas = this.canvas;
		const canvasContext = this.canvasContext;

		let ratio =  Math.max(window.devicePixelRatio || 1, 1);
		canvas.width = canvas.offsetWidth * ratio;
		canvas.height = canvas.offsetHeight * ratio;
		canvasContext.scale(ratio, ratio);
		this.clear();

		this.actions.classList.add('disabled');
	}
}
